'''
Created on Apr. 24, 2019

@author: jldupont
'''
import unittest

from ..state import States, StateExists, NoDefaultStateExists
from ..fsm import Fsm, ReportErrors, MissingState


class TestException(unittest.TestCase):
    
    def testOne(self):
        
        try:
            result = True # something that won't raise
        except:
            result = False
        else:
            result = 'OK'

        assert result == 'OK'


class Test(unittest.TestCase):

    @staticmethod
    def action_a(): pass

    @staticmethod
    def action1(): pass

    @staticmethod
    def action2(): pass
    
    @staticmethod
    def whatever(): pass
    
    @staticmethod
    def default(c): return c

    def testAddSimple(self):
               
        s=States()
        s.add(Test.action_a, 'a')
        
        assert s.get('a') == Test.action_a

    def testAddMissing(self):
        
        s=States()
        
        s.add(Test.action_a, 'a1', 'a2', 'a3')
        
        assert s.get('a1', 'a2', 'a3') == Test.action_a
        assert s.get('a1') == None

    def testAddComplex2(self):
        
        s=States()
        
        s.add(Test.action1, 'a11', 'a12', 'a13')
        s.add(Test.action2, 'a21', 'a22', 'a23')        
        
        assert s.get('a11', 'a12', 'a13') == Test.action1
        assert s.get('a21', 'a22', 'a23') == Test.action2
        
        assert s.get('a11') == None
        assert s.get('a11','a12') == None
        assert s.get('a11','a12','a*') == None

    def testDefault(self):
        
        s=States()
        s.add(Test.action1, 'a21', 'a22', 'a23')
        s.add(Test.default, 'a21', 'a22', '*')        
        
        result = s.step({'result': 666}, 'a21', 'a22', 'aX3')
        assert result['context']['result'] == 666

    def testAddRaiseStateExists1(self):
        
        s=States()
        
        s.add(Test.action_a, 'a1', 'a2', 'a3')

        self.assertRaises(StateExists, lambda: s.add(Test.whatever, 'a1', 'a2'))

    def testAddRaiseStateExists2(self):
        
        s=States()
        
        s.add(Test.action_a, 'a1', 'a2', 'a3')

        self.assertRaises(StateExists, lambda: s.add(Test.whatever, 'a1', 'a2', 'a3'))


"""
class TestReportErrors(unittest.TestCase):
    
    def testCallable(self):
        assert callable(ReportErrors.state_missing)

class TestFsm(unittest.TestCase):
    
    def testEcho(self):
        s=States() 
        f=Fsm(s)
        
        s.add(f.echo, 'c1')       
        
        ctx = {"in": "test"}
        
        result = f.step(ctx, "c1")
        assert result['context']['in'] == 'test'
        
    def testMissingState(self):
        s=States()
        f=Fsm(s)
        
        ctx={}
        self.assertRaises(MissingState, lambda: f.step(ctx,'invalidState'))
"""
 
if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on Apr. 28, 2019

@author: jldupont

Performs various operations on NL JSON stdin stream

'''
import sys
import click
import json

@click.command()
@click.option('--ignore', '-i', help="Ignore error", default=False, is_flag=True)
@click.option('--stderr', help="Errors will be printed to sys.stderr", is_flag=True)
@click.option('--keep', '-k', multiple=True, help="Key to keep in input object")
def command(ignore, stderr, keep):
    
    for line in sys.stdin:
        
        line = " ".join(line.split())
        
        if len(line) == 0:
            continue
        
        try:
            jobj = json.loads(line)
            
        except:
            jobj = None
            
            if stderr:
                print("Expecting JSON object, got:", line, file=sys.stderr)
            
            if ignore == False:
                raise Exception("Expecting JSON object")
            

        if jobj is None:
            continue

        out = {}

        for item in keep:
            if item in jobj:
                out[item] = jobj[item]
                
        print(json.dumps(out))
    

if __name__ == '__main__':
    command()


'''
Definition of the state machine for this utility

Created on Apr. 25, 2019
@author: jldupont


Example:
{ "type": "Feature", "properties": { "PCA_ID": 180000059, "POSTALCODE": "P0R1L0", "PROV": "ON", "MAF_ID
": 350002523, "PREC_CODE": 2, "PCA_COUNT": 516, "DOM_PCA": 0, "MULTI_PC": 0, "DEL_M_ID": "LB0001", "LON
GITUDE": -83.442568069900005, "LATITUDE": 46.289068144200002 }, "geometry": { "type": "MultiPolygon", "
coordinates": [ [ [ [ -83.432379699999956, 46.29642750000005 ], ... , [ -83.433095599999945, 46
.296466300000077 ], [ -83.432379699999956, 46.29642750000005 ] ] ] ] } },
'''
import json

class MissingEventHandling(Exception): pass
class InvalidStringForJsonObject(Exception): pass
class MissingState(Exception): pass
class ControlFlow(Exception): pass
class Skip(ControlFlow): pass
class EndOfFile(ControlFlow): pass


class Fsm:
    
    def __init__(self):
       
        self.previous_event = None
        self.current_prefix = None
        self.current_event = None
        self.feature_string = ""
    
    def submitEvent(self, prefix, event, value=None):
        """
        Dispatch event to the method
        that handles current state
        """
        self.previous_event = self.current_event
        
        try:
            self.current_prefix = prefix
            self.current_event = event
            self.current_value = value
            fnc = getattr(self, "state_"+self.state)
        except:
            raise MissingState("Missing state method: "+self.state)
            
        try:
            fnc(prefix, event, value)
        except Skip:
            pass
        
    def wait(self, expected_prefix, expected_event, *_):
        """
        Skips the current event if it does not pass the gate
        """
        if expected_prefix != self.current_prefix:
            raise Skip()
        
        if expected_event is not None:
            if expected_event != self.current_event:
                raise Skip()

    def accumulate(self):
        """
        The event data is kept at the instance level
        """
        fnc = getattr(self, "evt_"+self.current_event)
        if fnc is None:
            raise MissingEventHandling("Expected: "+str(self.current_event))

        fnc()

    def start_feature(self):
        self.feature_string = ""

    def evt_null(self):
        self.feature_string += "null"

    def evt_start_map(self):
        self.feature_string += "{ "

    def evt_end_map(self):
        self.feature_string += " }"

    def evt_map_key(self):
        self.feature_string += '"' + self.current_value + '":'
        
    def evt_start_array(self):
        self.feature_string += '['
        
    def evt_end_array(self):
        self.feature_string += ']'
        
    def evt_string(self):
        self.feature_string += '"' + self.current_value + '"'
        
    def evt_number(self):
        self.feature_string += str(self.current_value)

    def add_comma(self):
        self.feature_string += ','

    def maybe_add_comma(self):
        
        if self.previous_event == 'end_map' or self.previous_event == 'end_array':
            if self.current_event != 'end_map' and self.current_event != 'end_array':
                return self.add_comma()
        
        if self.previous_event == 'string' or self.previous_event == 'number':
            if self.current_event != 'end_map' and self.current_event != 'end_array':
                return self.add_comma()
                
        if self.current_event == 'end_map':
            return
        
        if self.current_event == 'end_array':
            return
        
        if self.previous_event is None or len(self.previous_event) == 0:
            return
        
        if self.previous_event == 'end_map':
            return 
        
        if self.previous_event == 'end_array':
            return
        
        if self.previous_event == 'start_map':
            return
        
        if self.previous_event == 'start_array':
            return
        
        if self.previous_event == 'map_key':
            return
        
        self.feature_string += ','

    def event_feature(self):
        """
        Receives complete string representing 
        a feature
        """
        try:
            # We just need to make we have a valid json
            _ = json.loads(self.feature_string)
            
        except Exception as _e:
            print(self.feature_string)
            print(_e)
            raise InvalidStringForJsonObject()
        else:
            print(self.feature_string+"\n")

'''
When to prepend a comma:

>  NUMBER , NUMBER
>  STRING , NUMBER
>  STRING , STRING
>  STRING , {
>  STRING , [
>  NUMBER , {
>  NUMBER , [
>  ] , [
>  

When NOT to prepend a comma:
>  ___ [
>  ___ {
>  ] ]
>  } }
>  ] }
>  } ]
'''

        
class GeojsonFsm(Fsm):
    
    def __init__(self):
        super().__init__()
        self.state = 'waitStartFeaturesItem'
        
    def state_waitStartFeaturesItem(self, prefix, event, value):
        
        if event == 'end_map':
            if prefix == '':
                if value == None or value == '':
                    raise EndOfFile()
        
        self.wait('features.item', 'start_map')
        self.start_feature()
        self.evt_start_map()
        self.state = 'accumulateSymbolsUntilFeatureEnd'
        
    def state_accumulateSymbolsUntilFeatureEnd(self, prefix, event, _value):
        
        if event == 'end_map':
            self.evt_end_map()
            if prefix == 'features.item':
                self.event_feature()
                self.state = 'waitStartFeaturesItem'
        else:
            self.maybe_add_comma()
            self.accumulate()




'''
Created on Apr. 25, 2019

@author: jldupont
'''
import unittest

from geojson_converter.fsm import GeojsonFsm, MissingState

class Test(unittest.TestCase):

    def setUp(self):
        
        self.f = GeojsonFsm()

    def testFsmSimple(self):
        
        self.f.submitEvent(None, None, None)
        
        assert self.f.state == 'waitForMapKeyFeatures'
        
    def testFsmSimple2(self):
        
        self.f.submitEvent('features', 'start_array')
        assert self.f.state == 'waitStartFeaturesItem'

    def testFsmSimple3(self):

        self.f.submitEvent('features', 'start_array')
        self.f.submitEvent('features.item', 'start_map')

        

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testFsmSimple']
    unittest.main()


#!/usr/bin/env python3
'''
Created on Apr. 26, 2019

@author: jldupont
'''
__author__ = "Jean-Lou Dupont"
__version__ = "0.4"

DESC="""
Feature Collection geojson file to NL JSON of Feature objects
"""

from distutils.core import setup
from setuptools import find_packages


setup(name=         'jldgeo',
      version=      __version__,
      description=  'Geojson Feature objects lister',
      author=       __author__,
      author_email= 'jl@jldupont.com',
      url=          'https://github.com/jldupont/jldgeo',
      package_dir=  {'': "src",},
      packages=     find_packages("src"),
      scripts=      ['src/scripts/jldgeojson'
                     ,'src/scripts/jldjson'
                     ],
      zip_safe=False
      ,long_description=DESC
      ,install_requires=[ "click>=7.0" 
                         ,'ijson>=2.3'
                         ]
      )

#############################################

f=open("latest", "w")
f.write(str(__version__)+"\n")
f.close()
